use clap::Parser;
use core::panic;
use std::fs::{read, write};
use toml::{Table, Value};

mod target_config;
use target_config::*;

#[derive(Parser)]
struct Args {
    #[arg(short, long, default_value = "ngen.toml")]
    config_file: String,

    #[arg(short, long, default_value = "build.ninja")]
    output_file: String,
}

fn target_generator(config: &TargetConfig) -> String {
    let mut ret = String::from("\n");
    let name = &config.name;
    let outfile = &config.outfile;
    let compiler = &config.compiler;
    let compiler_flags = &config.compiler_flags.join(" ");
    let linker = &config.linker;
    let linker_flags = config.linker_flags.join(" ");
    let linker_libs = config.linker_libs.join(" ");
    let sources = &config.sources;

    ret.push_str(&format!(
        "\
# BEGIN TARGET {name}
rule cc_{name}
  deps = gcc
  depfile = $dep
  command = {compiler} {compiler_flags} -MD -MF $dep -o $out -c $in
  description = Building object $out
rule link_{name}
  command = {linker} {linker_flags} -o $out $in {linker_libs}
  description = Linking $out

build $builddir/{name}/obj: mkdir
build $builddir/{name}/dep: mkdir

"
    ));

    let mut object_list: Vec<String> = Vec::new();

    for s in sources {
        let new_file = s.replace("/", "-");
        let obj_name = format!("$builddir/{name}/obj/{new_file}.o");
        let dep_name = format!("$builddir/{name}/dep/{new_file}.o.d");
        ret.push_str(&format!(
            "build {obj_name}: cc_{name} {s}\n  dep = {dep_name}\n"
        ));
        object_list.push(obj_name);
    }

    ret.push_str(&format!("\nbuild $builddir/{name}/{outfile}: link_{name} "));
    ret.push_str(&object_list.join(" "));
    ret.push_str(&format!(
        " || $builddir/{name}/obj $builddir/{name}/dep\nbuild {name}: phony $builddir/{name}/{outfile}\n"
    ));
    if config.is_default {
        ret.push_str(&format!("\ndefault {name}\n"));
    }
    ret.push_str(&format!("# END TARGET {name}\n"));

    ret
}

fn main() {
    let args = Args::parse();
    let output_file = &args.output_file;
    let config_file_name = &args.config_file;

    let config_file_raw =
        read(config_file_name).unwrap_or_else(|e| panic!("fatal: could not read file: {e}"));
    let config_file_string = String::from_utf8_lossy(&config_file_raw);
    let parsed_config = toml::from_str::<Table>(&config_file_string)
        .unwrap_or_else(|e| panic!("fatal: could not parse config file: {e}"));

    let mut targets: Vec<TargetConfig> = Vec::new();
    targets.push(TargetConfig::new(&parsed_config, "noname", None));
    for k in parsed_config.keys() {
        if let Value::Table(t) = parsed_config.get(k).unwrap() {
            targets.push(TargetConfig::new(&t, k, Some(&targets[0])));
        }
    }

    let mut ninjafile = String::from(
        "\
# Generated by ngen. Do not modify by hand.

builddir = build

rule mkdir
  command = mkdir -p $out
  description = Creating directory $out

rule regen
  command = ngen -c $in -o $out
  generator = 1
  description = Regenerating $out

",
    );

    ninjafile.push_str(&format!(
        "build {output_file}: regen {config_file_name}\n  pool = console\n"
    ));

    for t in targets {
        ninjafile.push_str(&target_generator(&t));
    }

    write(output_file, ninjafile)
        .unwrap_or_else(|e| panic!("fatal: could not write ninja file: {e}"));
}
