use toml::{Table, Value};

#[derive(Debug, Clone)]
pub struct TargetConfig {
    pub name: String,
    pub outfile: String,
    pub compiler: String,
    pub compiler_flags: Vec<String>,
    pub linker: String,
    pub linker_flags: Vec<String>,
    pub linker_libs: Vec<String>,
    pub sources: Vec<String>,
    pub is_default: bool,
}

impl TargetConfig {
    fn get_default() -> Self {
        TargetConfig {
            name: String::from("noname"),
            outfile: String::from("a.out"),
            compiler: String::from("cc"),
            compiler_flags: Vec::new(),
            linker: String::from(""),
            linker_flags: Vec::new(),
            linker_libs: Vec::new(),
            sources: Vec::new(),
            is_default: false,
        }
    }
    pub fn new(table: &Table, name: &str, global_conf: Option<&Self>) -> Self {
        let mut ret: Self;
        let mut linker_set = false;

        if let Some(global) = global_conf {
            ret = global.clone();
        } else {
            ret = Self::get_default();
        }

        ret.name = String::from(name);
        ret.is_default = false;

        for k in table.keys() {
            match &*k.to_owned() {
                "outfile" => {
                    if let Value::String(s) = table.get(k).unwrap() {
                        ret.outfile = s.to_owned();
                    } else {
                        panic!("fatal: {k} invalid type, must be a string.");
                    }
                }
                "compiler" => {
                    if let Value::String(s) = table.get(k).unwrap() {
                        ret.compiler = s.to_owned();
                    } else {
                        panic!("fatal: {k} invalid type, must be a string.");
                    }
                }
                "compiler_flags" => {
                    if let Value::Array(a) = table.get(k).unwrap() {
                        for element in a {
                            if let Value::String(s) = element {
                                ret.compiler_flags.push(s.to_owned());
                            } else {
                                panic!("fatal: {k} invalid type, must be a array of strings.");
                            }
                        }
                    } else {
                        panic!("fatal: {k} invalid type, must be a array of strings.");
                    }
                }
                "linker" => {
                    if let Value::String(s) = table.get(k).unwrap() {
                        ret.linker = s.to_owned();
                        linker_set = true;
                    } else {
                        panic!("fatal: {k} invalid type, must be a string.");
                    }
                }
                "linker_flags" => {
                    if let Value::Array(a) = table.get(k).unwrap() {
                        for element in a {
                            if let Value::String(s) = element {
                                ret.linker_flags.push(s.to_owned());
                            } else {
                                panic!("fatal: {k} invalid type, must be a array of strings.");
                            }
                        }
                    } else {
                        panic!("fatal: {k} invalid type, must be a array of strings.");
                    }
                }
                "linker_libs" => {
                    if let Value::Array(a) = table.get(k).unwrap() {
                        for element in a {
                            if let Value::String(s) = element {
                                ret.linker_libs.push(s.to_owned());
                            } else {
                                panic!("fatal: {k} invalid type, must be a array of strings.");
                            }
                        }
                    } else {
                        panic!("fatal: {k} invalid type, must be a array of strings.");
                    }
                }
                "sources" => {
                    if let Value::Array(a) = table.get(k).unwrap() {
                        for element in a {
                            if let Value::String(s) = element {
                                ret.sources.push(s.to_owned());
                            } else {
                                panic!("fatal: {k} invalid type, must be a array of strings.");
                            }
                        }
                    } else {
                        panic!("fatal: {k} invalid type, must be a array of strings.");
                    }
                }
                "default" => {
                    if let Value::Boolean(b) = table.get(k).unwrap() {
                        ret.is_default = *b;
                    } else {
                        panic!("fatal: {k} invalid type, must be a boolean.");
                    }
                }
                _ => {
                    if let Value::Table(_) = table.get(k).unwrap() {
                    } else {
                        panic!("fatal: unrecognized key {k}.");
                    }
                }
            }
        }

        if !linker_set {
            ret.linker = ret.compiler.clone();
        }

        if ret.sources == Vec::<String>::new() {
            panic!("fatal: you MUST specify at least one source");
        }

        ret
    }
}
